#!/usr/bin/env ruby
###############################################################
#   
#  File:      cli.rb
#
#  Author:    Christian Hättich
#
#  Project:   System-On-Chip Maker
#
#  Target:    Linux / Windows / Mac
#
#  Language:  ruby
#
#
###############################################################
#
#
#   Copyright (C) 2014  Christian Hättich  - feddischson [ at ] opencores.org
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
###############################################################
require 'readline'
require 'optparse'
require 'singleton'

module SOCMaker

#
# Command-line interface for accessing the SOC-Maker functionallity
# 
# PLEASE NOTE: this class is outdated and needs some work!!
# 
# The following commands are available:
# - new           -> create a new soc file
# - open          -> open soc file
# - list          -> list library
# - add           -> add core to the SOC
# - add_interface -> add an interface to the SOC
# - parameter     -> set/get parameter
# - sparameter    -> set/get static parameter
# - connect       -> connect cores
# - delete        -> delete core or connection
# - save          -> save soc
# - generate      -> generate soc
# - quit          -> quit this CLI
# - exit          -> same than quit
# - help          -> print some help
# 
# Please use the help command to get more information about
# each command and its parameters.
# 
# This CLI is a wrapper around SOCMaker::SOCDef and
# this class is realized as singleton.
#
# TODO: add commands for
# - selecting the coder (at the moment, only VHDL is supported)
# - refreshing the lib
#
class Cli

  # to organize it as singleton
  #
  include Singleton


  #
  # the SOC, which is edited by this script
  # 
  @soc      = nil


  #
  # An abbreviation map.
  #
  @appr_map = {}


  #
  # All available commands
  #
  @commands = []

  # 
  # Constructor of this CLI: it sets up a abbreviation map and a list
  # of available commands. Moreover, it initializes readline.
  #
  def initialize

    # abbreviation map
    @appr_map = { 'n' => "new",
                  'o' => "open",
                  'q' => "quit",
                  'h' => "help",
                  'l' => "list", 
                  'a' => "add",
                  'g' => "generate",
                  's' => "save",
                  'p' => "parameter",
                  'd' => "delete",
                  'c' => "connect",
                  'i' => "print",
                  'x' => "exit"
                  }

    # all available commands
    @commands = %w[ new open list add add_interface parameter sparameter  
                    delete connect save help quit exit 
                    generate print set get ]

    comp = proc { |s| (@commands + Dir.entries( Dir.pwd )).grep( /^#{Regexp.escape(s)}/ ) }
    Readline.completion_append_character = " "
    Readline.completion_proc = comp

  end

  #
  # Method to start processing the commands. Readline us used
  # to receive input data. Each line to passed to process_cmd.
  #
  def run

    ##
    # process user commands
    #
    while buf = Readline.readline( "> ", true )
      process_cmd buf
    end
  end

  #
  # Method to processes a single command
  #
  def process_cmd( c )

      # remove the comments and split each line
      match = SOCMaker::conf[ :COMMENT_REGEX ].match( c )
      cmd_arr = match[1].split( ' ' )

      # process the command, if there is one
      if cmd_arr.size > 0 
        cmd     = ""
        if cmd_arr[ 0 ].size == 1 and @appr_map[ cmd_arr[ 0 ] ] != nil
          cmd = @appr_map[ cmd_arr[ 0 ] ] 
        else
          cmd = cmd_arr[ 0 ] 
        end

        if @commands.include?( cmd )
          cmd_str = "do_#{cmd}( cmd_arr[ 1..-1] )"  
          puts "evaluating >>#{cmd_str}<< "
          eval( cmd_str ) 
        elsif !Gem.win_platform? && system( "which #{cmd} > /dev/null 2>&1" )
          # this is for linux only
          system( c )
        else
          puts "Command #{cmd} not available"
        end
       #begin
       #rescue 
       #  puts "evaluating >>#{cmd_str}<< failed"
       #end
      end
  end


private


  # failed message
  FMSG      = ' -> failed '


  #########################################
  #
  # command implementations:
  #   -> a usage string and
  #      a function for every command
  #

  #
  # Usage message of the 'new' comand
  #
  NEW_USAGE = 
  "  > new <<name>> <<id>> <<toplevel>>   # opens a system-on-chip file
         - <<name>>     : the SOC name
         - <<id>>       : the SOC id
         - <<toplevel>> : the toplevel name
  "

  #
  # Implementation of the 'new' command
  #
  def do_new( args )
    if args.size != 3
      puts "three arguments are required:\nusage:\n#{NEW_USAGE}"
    else
      @soc = SOCMaker::SOCDef.new( args[0], args[1], args[2] ) 
      SOCMaker::lib.add_core( @soc )
      @soc.consistence_check
    end
  end

  #
  # Usage message of the 'open' comand
  #
  OPEN_USAGE =
  "  > open <<file>>    # opens a system-on-chip file
         - <<file>>    : system-on-chip definition in in YAML format

  "
  #
  # Implementation of the 'open' command
  #
  def do_open( args )
    if args.size != 1
      puts "only one argument is required:\nusage:\n#{OPEN_USAGE}"
    else
      puts "loading #{args[0]}"
      @soc = SOCMaker::from_f( args[0] ) 
      SOCMaker::lib.add_core( @soc )
      @soc.consistence_check
    end
  end

  #
  # Usage message of the 'list' comand
  #
  LIST_USAGE = 
  "  > list             # prints list of cores and interfaces,
                          which are in the library

  "
  #
  # Implementation of the 'list' command
  #
  def do_list( args )
    puts SOCMaker::lib
  end

  #
  # Usage message of the 'add' comand
  #
  ADD_USAGE = 
  "  > add <<id>> <<name>>    
                        # adds an ip-core from the library to the SOC
        - <<id>>      : id of the IP core
        - <<name>>    : instanciation name

  "
  #
  # Implementation of the 'add' command
  #
  def do_add( args )
    if args.size != 2
      puts "two arguments are required:\nusage:\n#{ADD_USAGE}"
    else
      @soc.add_core( args[ 0 ].to_sym, args[ 1 ].to_sym ) 
      @soc.consistence_check
    end
  end


  ADD_INTERFACE_USAGE =
  "  > add_interface  <<name>> <<id>> <<dir>> <<port_name>> <<port_size>> [ <<port_name>> <<port_size>> .... ]
                        # adds an interface to the SOC
        - <<name>>      name of the interface
        - <<id>>        id of the interface specification
        - <<dir>>       the direction
        - <<port_name>> name of the port
        - <<port_size>> size of the port

  "
  def do_add_interface( args )


    if args.size < 5 || args.size % 2 != 1
      puts "five or more (but odd) arguments required:\nusage:\n#{ADD_INTERFACE_USAGE}"
    else
      ports = {}
      args[3..-1].each_slice(2).to_a.each do |entry| 
        ports[ entry[0].to_sym ] = SOCMaker::IfcPort.new( entry[0], entry[1] )
      end

      @soc.add_interface( args[0], args[ 1 ], args[2].to_i, ports )

    end

  end

  #
  # Usage message of the 'parameter' comand
  #
  PARAMETER_USAGE = 
  "  > prameter <<instance>> <<parameter>> <<value>>
                        # modifies a parameter of an instance
       - <<instance>>   : the instance name of the core
       - <<parameter>>  : the instance parameter name 
       - <<value>>      : the value which is set (optional). The current 
                          value is printed, if omitted
  "
  #
  # Implementation of the 'parameter' command
  #
  def do_parameter( args )
    if args.size == 2 
      @soc.get_param( args[ 0 ].to_sym, args[ 1 ].to_sym ) 
    elsif args.size == 3
      @soc.set_param( args[ 0 ].to_sym, args[ 1 ].to_sym, args[ 2 ] ) 
      @soc.consistence_check
    else
      puts "two or three arguments required:\nusage:\n#{PARAMETER_USAGE}"
    end
  end

  #
  # Usage message of the 'sparameter' comand
  #
  SPARAMETER_USAGE = 
  "  > sprameter <<core>> <<parameter>> <<value>>
                        # modifies the static parameter of a core
       - <<core>>       : the name of the core
       - <<parameter>>  : the static parameter name 
       - <<value>>      : the value which is set (optional). The current 
                          value is printed, if omitted
  "
  #
  # Implementation of the 'sparameter' command
  #
  def do_sparameter( args )
    if args.size == 2 
      @soc.get_sparam( args[ 0 ].to_sym, args[ 1 ].to_sym )
    elsif args.size == 3
      @soc.set_sparam( args[ 0 ].to_sym, args[ 1 ].to_sym, args[ 2 ] ) 
      @soc.consistence_check
    else
      puts "two or three arguments required:\nusage:\n#{SPARAMETER_USAGE}"
    end
  end

  #
  # Usage message of the 'connect' comand
  #
  CONNECT_USAGE = 
  "  > connect <<name>> <<core1>> <<ifc1>> <<core2>> <<ifc2>> 
                        # connects two cores
        - <<name>>      : connection name
        - <<core1>>     : instance name of the first core
        - <<core2>>     : instance name of the second core
        - <<ifc1>>      : interface name of the first core
        - <<ifc2>>      : interface name of the second core

  "      
  #
  # Implementation of the 'connect' command
  #
  def do_connect( args )
    if args.size != 5
      puts "five arguments are required:\nusage:\n#{CONNECT_USAGE}"
    else
      @soc.add_connection( args[ 0 ], 
                           args[ 1 ].to_sym, 
                           args[ 2 ].to_sym, 
                           args[ 3 ].to_sym, 
                                        args[ 4 ].to_sym )
      @soc.consistence_check
    end
  end

  #
  # Usage message of the 'delete' comand
  #
  DELETE_USAGE = 
  "  > delete <<core/connection>>
                        # removes a core or a connection
        - <<core/conection> : the core or connection, which is removed

  "
  #
  # Implementation of the 'delete' command
  #
  def do_delete( args )
    if args.size != 1 
      puts "five arguments are required:\nusage:\n#{DELETE_USAGE}"
    else
      @soc.rm( args[ 0 ].to_sym ) 
      @soc.consistence_check
    end
  end
  #
  # Usage message of the 'save' comand
  #
  SAVE_USAGE = 
  "  > save <<file>>    # saves system-on-chip definition in YAML format to file
        - <<file>>     : optional destination file, when omitted: the 
                         original file-path is used

  "
  #
  # Implementation of the 'save' command
  #
  def do_save( args )
    if args.size > 1
      puts "zero or one argument is required:\nusage:\n#{SAVE_USAGE}"
    else
      p args
      @soc.save_yaml( args ) 
      File.open( args[0], 'w' ) { |f| f.write( SOCMaker::to_yaml_s( @soc )  ) }
    end
  end

  #
  # Usage message of the 'generate' comand
  #
  GENERATE_USAGE = 
  "  > generate         # generates a synthesizable system-on-chip implementation

  "
  #
  # Implementation of the 'generate' command
  #
  def do_generate( args )
    if args.size != 0
      puts "no arguments are required:\nusage:\n#{GENERATE_USAGE}"
    else
      @soc.consistence_check
      SOCMaker::deploy_soc( @soc )
    end
  end

  #
  # Usage message of the 'print' comand
  #
  PRINT_USAGE = 
  "  > print            # prints SOC information

  "
  #
  # Implementation of the 'print' command
  #
  def do_print( args )
    if args.size != 0
      puts "no arguments are required:\nusage:\n#{PRINT_USAGE}"
    else
      puts @soc
    end
  end

  #
  # Usage message of the 'quit' comand
  #
  QUIT_USAGE = 
  "  > quit             # the same than exit

  "
  #
  # Implementation of the 'quit' command
  #
  def do_quit( args )
    do_exit( args )
  end

  #
  # Usage message of the 'exit' comand
  #
  EXIT_USAGE = 
  "  > exit             # exits this tool

  "
  #
  # Implementation of the 'exit' command
  #
  def do_exit( args )
    puts "... bye bye!"
    exit 0
  end

  #
  # Usage message of the 'help' comand
  #
  HELP_USAGE = 
  "  > help             # prints some help information

  "
  #
  # Implementation of the 'help' command
  #
  def do_help( args )
    puts "The following commands are available:\n\n"
    @commands.each { |c| eval "puts  #{c.upcase}_USAGE" }
  end

  #
  # Usage message of the 'set' comand
  #
  SET_USAGE = 
  "  > set              # not implemented yet

  "
  #
  # Implementation of the 'set' command
  #
  def do_set( args )
    puts "NOT IMPLEMENTED, YET"
  end
  #
  # Usage message of the 'get' comand
  #
  GET_USAGE = 
  "  > get              # not implemented yet

  "
  #
  # Implementation of the 'get' command
  #
  def do_get( args )
    puts "NOT IMPLEMENTED, YET"
  end

  #  end command implementations
  #
  #################################




end
end


# vim: noai:ts=2:sw=2

