###############################################################
#   
#  File:      err.rb
#
#  Author:    Christian Hättich
#
#  Project:   System-On-Chip Maker
#
#  Target:    Linux / Windows / Mac
#
#  Language:  ruby
#
#
###############################################################
#
#
#   Copyright (C) 2014  Christian Hättich  - feddischson [ at ] opencores.org
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
###############################################################
module SOCMaker

  #
  # This sub-module contains some error-functionallity,
  # which is used in different classes via mixins.
  #
  module ERR

    #
    # An error, which is thrown if YAML parsing fails
    #
    class YAMLParseError < RuntimeError
    end



    #
    # Processing errors: occur during the the processing
    # of SoCs.
    # 
    class ProcessingError < RuntimeError
      #
      # Optional error information
      #
      attr :optinal
      #
      # This constructor saves the message and optional infos and 
      # adds a log entry via SOCMaker::logger.error
      # +message+:: the error message
      # +o+:: optional error information
      #
      def initialize( message, o={} )
        super message
        @optional = o
        SOCMaker::logger.error( "Processing Error Raised: #{message}  #{ "\n\t\t" + o.inspect if o.size > 0}" )
      end
      #
      # Returns a string describing this error
      #
      def to_s
        "#{super} #{ "\n\t\t" + @optional.inspect if @optional.size > 0}"
      end
    end

    #
    # Method to raise a processing error
    # +msg+:: error message
    # +o+:: optional info
    #
    def processing_error( msg, o={} )
      raise ProcessingError.new( msg, o )
    end

    # Method to raise a processing error
    # if the condition is true
    # +con+:: condition
    # +msg+:: error message
    # +o+:: optional info
    def processing_error_if( res, msg, o={} )
      if !!(res)
        processing_error( msg, o )
      end
    end






    #
    # Consistence errors: occur during the consistence check
    # of the classes or in a few methods, where a inconsistence
    # can be caused by external call.
    # A cause for this error could be, that a core or interface
    # doesn't exist in the library. Or a port in of a IfcDef is
    # not defined in the IfcSpc.
    #
    class ConsistenceError < RuntimeError
      #
      # Optional error information
      #
      attr :optinal
      #
      # This constructor saves the message and optional infos and 
      # adds a log entry via SOCMaker::logger.error
      # +message+:: the error message
      # +o+:: optional error information
      #
      def initialize( message, o={} )
        super message
        @optional = o
        SOCMaker::logger.error( "Consistence Error Raised: #{message}  #{ "\n\t\t" + o.inspect if o.size > 0}" )
      end
      #
      # Returns a string describing this error
      #
      def to_s
        "#{super} #{ "\n\t\t" + @optional.inspect if @optional.size > 0}"
      end
    end

    #
    # Method to raise a consistence error
    # +msg+:: error message
    # +o+:: optional info
    #
    def consistence_error( msg, o={} )
      raise ConsistenceError.new( msg, o )
    end

    # Method to raise a consistence error
    # if the condition is true
    # +con+:: condition
    # +msg+:: error message
    # +o+:: optional info
    def consistence_error_if( con, msg, o={} )
      if !!(con)
        consistence_error( msg, o )
      end
    end



    #
    # Initialization errors: occur during the initialization
    # of the classes. A cause for this error could be a wrong structure 
    # of the YAML file, for example wrong data types.
    #
    class InitError < RuntimeError
      #
      # Optional error information
      #
      attr :optinal
      #
      # This constructor saves the message and optional infos and 
      # adds a log entry via SOCMaker::logger.error
      # +message+:: the error message
      # +o+:: optional error information
      #
      def initialize( message, o={} )
        super message
        @optional = o
        if SOCMaker::logger
          SOCMaker::logger.error( "Initialization Error Raised: #{message}  #{ "\n\t\t" + o.inspect if o.size > 0}" )
        end
      end
      #
      # Returns a string describing this error
      #
      def to_s
        "#{super} #{ "\n\t\t" + @optional.inspect if @optional.size > 0}"
      end
    end

    #
    # Method to raise an initialization error
    # +msg+:: error message
    # +o+:: optional info
    #
    def init_error( msg, o={} )
      raise InitError.new( msg, o )
    end

    # Method to raise an initialization error
    # if the condition is true
    # +con+:: condition
    # +msg+:: error message
    # +o+:: optional info
    def init_error_if( con, msg, o={} )
      if !!(con)
        init_error( msg, o )
      end
    end


  end # module ERR

end # module SOCMaker
