###############################################################
#   
#  File:      ifc_port.rb
#
#  Author:    Christian Hättich
#
#  Project:   System-On-Chip Maker
#
#  Target:    Linux / Windows / Mac
#
#  Language:  ruby
#
#
###############################################################
#
#
#   Copyright (C) 2014  Christian Hättich  - feddischson [ at ] opencores.org
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
###############################################################
#
#
###############

module SOCMaker



#
# A small classes, used to group information
# and to verify, auto-correct and auto-complete
# this information:
# The class represents an interface port within 
# SOCMaker::IfcDef.
# It is used to make a relation between the naming
# of SOCMaker::IfcDef and SOCMaker::IfcSpc.
# The two data fiels are
# - #spc_ref (mandatory: must be the name of the port, defined in IfcSpc)
# - #len (optional, default is 1)
#
class IfcPort 
  include ERR

  #
  # Used as reference to a port in SOCMaker::IfcSpc
  #
  attr_accessor :spc_ref

  #
  # Length of this port
  #
  attr_accessor :len


  #
  # Constructor, which expects #spc_ref and #len. Length has the default value 1.
  #
  def initialize( spc_ref, len = 1 )
    init_with( 'spc_ref' => spc_ref,
               'len'  => len )
  end


  # 
  # Encoder method (to yaml)
  #
  # +coder+:: An instance of the Psych::Coder to encode this class to a YAML file
  #
  def encode_with( coder )

    init_error_if !coder.is_a?( Psych::Coder ), 
                'coder is not given as Psych::Coder'
    %w[ spc_ref len ].
          each { |v| coder[ v ] = instance_variable_get "@#{v}" }
  end

  #
  # Initialization method (from yaml)
  #
  # +coder+:: An instance of the Psych::Coder to init this class from a YAML file
  #
  #
  def init_with( coder )

    init_error_if !( coder.is_a?( Hash ) || coder.is_a?( Psych::Coder ) ), 
                'coder is not given as Hash neither as Psych::Coder'

    init_error 'no relation to interface-definition is given for an interface port (nil)',
      field: "spc_ref" if coder[ 'spc_ref' ] == nil
    @spc_ref = coder[ 'spc_ref' ]

    init_error 'Relation to interface definition is not of type string',
      instance: @spc_ref.to_s,
      field:    "spc_ref" if  !@spc_ref.is_a?( String )

    init_error 'Relation to interface definition has zero length',
      instance: @spc_ref.to_s,
      field:    "spc_ref" if @spc_ref.size == 0

    @len = coder[ 'len' ] || 1

    init_error 'Length is not a fixnum',
      instance: @spc_ref.to_s,
      field:    "spc_ref" if !( @len.is_a?( Fixnum ) || @len.is_a?( String ) )
  end



  #
  # Equality operator
  #
  def ==(o)
    o.class   == self.class     && 
    o.spc_ref == self.spc_ref   &&
    o.len     == self.len 
  end 


end
end
