###############################################################
#   
#  File:      parameter.rb
#
#  Author:    Christian Hättich
#
#  Project:   System-On-Chip Maker
#
#  Target:    Linux / Windows / Mac
#
#  Language:  ruby
#
#
###############################################################
#
#
#   Copyright (C) 2014  Christian Hättich  - feddischson [ at ] opencores.org
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
###############################################################


module SOCMaker
#
#
# A small classes, used to group information
# and to verify, auto-correct and auto-complete
# this information:
# This class represents an instance parameter for 
# a core with the following values:
# - type (mandatory)
# - default
# - min
# - max 
# - visible 
# - editable
# - description
# Most of the fields are reserved for future implementations.
#
class Parameter 
  include ERR

  # type of the parameter
  attr_accessor :type

  # default value
  attr_accessor :default

  # min value
  attr_accessor :min

  # max value
  attr_accessor :max

  # visibility
  attr_accessor :visible

  # editable flag
  attr_accessor :editable

  # description
  attr_accessor :description

  # choice: if type == 'enum', choice must be an array of choices
  attr_accessor :choice

  #
  # This constructor expects only the type as mandatory parameter, 
  # everything else can be passed as optional parameter.
  #
  def initialize( type, optional = {} )
    init_with( { 'type' => type }.merge( optional ) )
  end

  # 
  # Encoder method (to yaml)
  #
  # +coder+:: An instance of the Psych::Coder to encode this class to a YAML file
  #
  def encode_with( coder )
    init_error_if !coder.is_a?( Psych::Coder ), 
                'coder is not given as Psych::Coder'
    %w[ type default min max
        visible editable description ].
          each { |v| coder[ v ] = instance_variable_get "@#{v}" }
  end

  #
  # Initialization method (from yaml)
  #
  # +coder+:: An instance of the Psych::Coder to init this class from a YAML file
  #
  #
  def init_with( coder )
    init_error_if !( coder.is_a?( Hash ) || coder.is_a?( Psych::Coder ) ), 
                'coder is not given as Hash neither as Psych::Coder'

    init_error 'no parameter type specified',
      field: "type" if coder[ 'type' ] == nil

    @type = coder[ 'type' ]
    init_error "Parameter type is not defined with string",
      field: "parameter" if !@type.is_a?( String )

    init_error "Parameter type string has zero length",
      field: "parameter" if @type.size == 0

    @default      = coder[ 'default'     ] || 0
    @min          = coder[ 'min'         ] || 0
    @max          = coder[ 'max'         ] || 0
    @visible      = coder[ 'visible'     ] || true
    @editable     = coder[ 'editable'    ] || false
    @description  = coder[ 'description' ] || ''
    @choice       = coder[ 'choice'      ] || []
  end


  #
  # Equality operator
  #
  def ==(o)
    o.class         == self.class       && 
      o.type        == self.type        &&
      o.default     == self.default     &&
      o.min         == self.min         &&
      o.max         == self.max         &&
      o.visible     == self.visible     &&
      o.editable    == self.editable    &&
      o.description == self.description &&
      o.choice      == self.choice
  end


end
end

